#!/usr/bin/env python2.3

import gzip
import re
import string

def transExtract(a, ind):
    if len(a) > ind:
        return a[ind]
    else:
        return None
    
def transpose(A):
    ret = [[transExtract(a, ind) for a in A] for ind in range(len(A[0]))]
    for ind, row in enumerate(ret): 
        if row == [None] * len(row):
            ret[ind] = []
    return ret

def valueType(a):
    if type(a) == type(1):
        return 30
    elif type(a) == type(1.0):
        return 40
    elif type(a) == type(""):
        return 60
    return 0

def findSubStrIndex(data, tofind, start, end):
    for ind in range(start, min(end, len(data))):
        if tofind in data[ind]:
            return ind
    raise IndexError

def hasResultsSheet(filename, resultName):
    # Get the data
    data = gzip.open(filename).readlines()

    tofind = "<gmr:SheetName>" + resultName + "</gmr:SheetName>"
    try:
        findSubStrIndex(data, tofind, 0, len(data))
        return 1
    except:
        return 0

def sheetNames(filename):
    data = gzip.open(filename).readlines()
    ret = []
    for line in data:
        if "SheetName" in line and "SheetNameIndex" not in line:
            line = string.strip(line)
            line = line.replace("<gmr:SheetName>", "")
            line = line.replace("</gmr:SheetName>", "")
            ret.append(line)
    return ret

def findDuplicateRows(filename, results, sheetName):
    data = gzip.open(filename).readlines()
    
    sheetS = findSubStrIndex(data, "<gmr:Name>" + sheetName + "</gmr:Name>",
                             0, len(data))
    endInd = findSubStrIndex(data, "</gmr:Sheet>", sheetS, len(data))

    dataStart = findSubStrIndex(data, "<gmr:Cells>", sheetS, endInd) + 1
    dataEnd = findSubStrIndex(data, "</gmr:Cells>", dataStart, endInd)

    ret = []
    names = {}
    for row in results:
        if row:
            names[row[0]] = 1

    for line in data[dataStart:dataEnd]:
        (col, row, type, value) = extractCellInfo(line)
        if col == 0 and row != 0:
            if value in names:
                ret.append(value)

    return ret

def columnNames(filename, sheetName):
    data = gzip.open(filename).readlines()
    ret = []

    sheetS = findSubStrIndex(data, "<gmr:Name>" + sheetName + "</gmr:Name>",
                             0, len(data))
    endInd = findSubStrIndex(data, "</gmr:Sheet>", sheetS, len(data))

    dataStart = findSubStrIndex(data, "<gmr:Cells>", sheetS, endInd) + 1
    dataEnd = findSubStrIndex(data, "</gmr:Cells>", dataStart, endInd)

    pattern = re.compile('.*Row="0"')
    for line in data[dataStart:dataEnd]:
        m = pattern.match(line)
        if not m:
            continue
        
        (col, row, type, value) = extractCellInfo(line)
        if col >= len(ret):
            ret += [None] * (col - len(ret) + 1)
        ret[col] = value
        
    return ret


def selectSheet(filename, sheetInd):
    data = gzip.open(filename).readlines()

    ind = findSubStrIndex(data, "<gmr:UIData SelectedTab",
                          len(data) - 50, len(data))

    data[ind] = '  <gmr:UIData SelectedTab="%d"/>\n' % sheetInd

    fh = gzip.open(filename, 'w')
    for line in data:
        print >>fh, line,
    fh.close()


def typed(value, valuetype):
    try:
        if valuetype == 30:
            return int(value)
        if valuetype == 40:
            return float(value)
        if valuetype == 60:
            return str(value)
    except:
        pass
    
    return None
        
def extractResultsSheet(filename, sheetName):
    data = gzip.open(filename).readlines()
    sheetS = findSubStrIndex(data, "<gmr:Name>" + sheetName + "</gmr:Name>",
                             0, len(data))
    endInd = findSubStrIndex(data, "</gmr:Sheet>", sheetS, len(data))

    dataStart = findSubStrIndex(data, "<gmr:Cells>", sheetS, endInd) + 1
    dataEnd = findSubStrIndex(data, "</gmr:Cells>", dataStart, endInd)

    results = []

    try:
        for ind in range(dataStart, dataEnd):
            (col, row, valuetype, value) = extractCellInfo(data[ind])
            if col == -1:
                continue

            if row >= len(results):
                for i in range(row - len(results) + 1):
                    results.append([])
                assert(len(results) == row + 1)

            if col >= len(results[row]):
                for i in range(col - len(results[row]) + 1):
                    results[row].append(None)
                assert(len(results[row]) == col + 1)

            results[row][col] = typed(value, valuetype)

    except:
        print results
        print row
        print col
        print len(results)
        raise
    
    for i in range(len(results) - 1, -1, -1):
        if results[i] == []:
            del results[i]

    return results

def addResultsSheet(filename, results, resultName):
    assert(not hasResultsSheet(filename, resultName))
    
    # Get the data
    data = gzip.open(filename).readlines()

    # Scan through the XML of this file to find SheetNameIndex
    tofind = "</gmr:SheetNameIndex>"
    ind = findSubStrIndex(data, tofind, 0, len(data))

    # Insert into the index.  This doesn't actually do a whole lot
    # except tell Gnumeric to look for a gmr:Sheet with that name
    data[ind - 1:ind - 1] = \
             ["    <gmr:SheetName>" + resultName + "</gmr:SheetName>\n"]

    # Scan through the file to find </gmr:Sheets>
    tofind = "</gmr:Sheets>"
    ind = findSubStrIndex(data, tofind, ind, len(data))

    # Insert the sheet data.
    data[ind: ind] = createSheet(results, resultName)
           
    # Save the file
    gzip.open(filename, 'w').writelines(data)

def getColumn(filename, sheetname, column):
    data = gzip.open(filename).readlines()
    
    tofind = "<gmr:Name>" + sheetname + "</gmr:Name>"
    ind = findSubStrIndex(data, tofind, 0, len(data))

    start = findSubStrIndex(data, "<gmr:Cells>", ind, len(data)) + 1
    end = findSubStrIndex(data, "</gmr:Cells>", start, len(data))

    ret = []
    for i in range(start, end):
        (col, row, valtype, value) = extractCellInfo(data[i])
        if col == column:
            while len(ret) - 1 < row: 
                ret.append([])

            ret[row] = value

    return ret

def overwriteResultsSheet(filename, results, resultName):    
    if not hasResultsSheet(filename, resultName):
        addResultsSheet(filename, results, resultName)
        return

    # Get the data
    data = gzip.open(filename).readlines()

    # Scan through the file to find the various sheets
    sheetStart = findSubStrIndex(data, "<gmr:Sheet ", 0, len(data)) + 1
    while 1:
        try:
            endInd = findSubStrIndex(data, "</gmr:Sheet", \
                                     sheetStart, len(data))
            nameInd = findSubStrIndex(data,
                                      "<gmr:Name>" + resultName + "</",
                                      sheetStart, endInd)
            break
        except IndexError:
            pass
        
        try:
            sheetStart = findSubStrIndex(data, "<gmr:Sheet ", \
                                         sheetStart, len(data)) + 1
        except IndexError:
            print "Gnumeric file corrupted, bailing."
            return

    sheetEnd = findSubStrIndex(data, "</gmr:Sheet", sheetStart, len(data)) + 1
    sheetStart -= 1

    # Insert the sheet data.
    data[sheetStart:sheetEnd] = createSheet(results, resultName)
           
    # Save the file
    gzip.open(filename, 'w').writelines(data)

def mergeResultsSheet(filename, results, resultName):    
    if not hasResultsSheet(filename, resultName):
        addResultsSheet(filename, results, resultName)
        return

    # Get the data
    data = gzip.open(filename).readlines()

    # Scan through the file to find the various sheets
    sheetStart = findSubStrIndex(data, "<gmr:Sheet ", 0, len(data)) + 1
    while 1:
        try:
            endInd = findSubStrIndex(data, "</gmr:Sheet", \
                                     sheetStart, len(data))
            nameInd = findSubStrIndex(data,
                                      "<gmr:Name>" + resultName + "</",
                                      sheetStart, endInd)
            break
        except IndexError:
            pass
        
        try:
            sheetStart = findSubStrIndex(data, "<gmr:Sheet ", \
                                         sheetStart, len(data)) + 1
        except IndexError:
            print "Gnumeric file corrupted, bailing."
            return

    sheetEnd = findSubStrIndex(data, "</gmr:Sheet", sheetStart, len(data)) + 1
    sheetStart -= 1

    # Now we need to merge our data into this sheet

    # Get the range for the cells
    cellsStart = findSubStrIndex(data, "<gmr:Cells>", sheetStart, sheetEnd)
    cellsEnd = findSubStrIndex(data, "</gmr:Cells>", cellsStart, sheetEnd)

    maxCol = 0
    for line in data[cellsStart:cellsEnd]:
        (col, row, type, value) = extractCellInfo(line)
        if col > maxCol:
            maxCol = col

    for row in range(len(results)):
        for col in range(len(results[row])):
            if results[row][col] == None:
                continue
            
            if str(results[row][col])[0] == '=':
                newCell = '<gmr:Cell Col="%d" Row="%d">%s</gmr:Cell>\n' % \
                          (col, row, str(results[row][col]))
            else:
                newCell = '<gmr:Cell Col="%d" Row="%d" ValueType="%d">%s' % \
                          (col, row,
                           valueType(results[row][col]),
                           str(results[row][col])) +\
                           '</gmr:Cell>\n'
            newCell = " " * 8 + newCell
            try:
                rowInd = findSubStrIndex(data,
                                         'Col="%d" Row="%d"' % (col, row),
                                         cellsStart, cellsEnd)
                # Can just update the data            
                data[rowInd] = newCell
                
            except IndexError:
                # Need to add this one
                data[cellsEnd:cellsEnd] = [newCell]
                cellsEnd += 1

        # In case there was MORE data on this row
        while col <= maxCol:
            try:
                rowInd = findSubStrIndex(data,
                                         'Col="%d" Row="%d"' % (col, row),
                                         cellsStart, cellsEnd)
                del data[rowInd]
                cellsEnd -= 1
            except IndexError:
                pass
            col += 1
            
    # Save the file
    gzip.open(filename, 'w').writelines(data)
    
def createSheet(results, resultName):
    data = []
    data.append('    <gmr:Sheet DisplayFormulas="false" HideZero="false" HideGrid="false" HideColHeader="false" HideRowHeader="false" DisplayOutlines="true" OutlineSymbolsBelow="true" OutlineSymbolsRight="true">\n')
    data.append("      <gmr:Name>" + resultName + "</gmr:Name>")
    data.append("      <gmr:MaxCol>" + str(len(results[0])) + "</gmr:MaxCol>")
    data.append("      <gmr:MaxRow>" + str(len(results)) + "</gmr:MaxRow>")

    sheetHeader = '''
      <gmr:Zoom>1.000000</gmr:Zoom>
      <gmr:Names/>
      <gmr:PrintInformation>
        <gmr:Margins>
          <gmr:top Points="28.3" PrefUnit="cm"/>
          <gmr:bottom Points="28.3" PrefUnit="cm"/>
          <gmr:left Points="28.3" PrefUnit="cm"/>
          <gmr:right Points="28.3" PrefUnit="cm"/>
          <gmr:header Points="14.2" PrefUnit="cm"/>
          <gmr:footer Points="14.2" PrefUnit="cm"/>
        </gmr:Margins>
        <gmr:Scale type="percentage" percentage="100"/>
        <gmr:vcenter value="0"/>
        <gmr:hcenter value="0"/>
        <gmr:grid value="0"/>
        <gmr:even_if_only_styles value="0"/>
        <gmr:monochrome value="0"/>
        <gmr:draft value="0"/>
        <gmr:titles value="0"/>
        <gmr:repeat_top value=""/>
        <gmr:repeat_left value=""/>
        <gmr:order>r_then_d</gmr:order>
        <gmr:orientation>landscape</gmr:orientation>
        <gmr:Header Left="" Middle="&amp;[TAB]" Right=""/>
        <gmr:Footer Left="" Middle="Page &amp;[PAGE]" Right=""/>
        <gmr:paper>A4</gmr:paper>
      </gmr:PrintInformation>
      <gmr:Styles>
        <gmr:StyleRegion startCol="0" startRow="0" endCol="255" endRow="65535">
          <gmr:Style HAlign="1" VAlign="2" WrapText="0" Orient="1" Shade="0" Indent="0" Locked="1" Hidden="0" Fore="0:0:0" Back="FFFF:FFFF:FFFF" PatternColor="0:0:0" Format="General">
            <gmr:Font Unit="9" Bold="0" Italic="0" Underline="0" StrikeThrough="0">Times</gmr:Font>
            <gmr:StyleBorder>
              <gmr:Top Style="0"/>
              <gmr:Bottom Style="0"/>
              <gmr:Left Style="0"/>
              <gmr:Right Style="0"/>
              <gmr:Diagonal Style="0"/>
              <gmr:Rev-Diagonal Style="0"/>
            </gmr:StyleBorder>
          </gmr:Style>
        </gmr:StyleRegion>
      </gmr:Styles>
      <gmr:Cols DefaultSizePts="48">
        <gmr:ColInfo No="0" Unit="70.5" MarginA="2" MarginB="2" HardSize="1"/>
        <gmr:ColInfo No="1" Unit="51" MarginA="2" MarginB="2" HardSize="1"/>
        <gmr:ColInfo No="2" Unit="54.8" MarginA="2" MarginB="2" HardSize="1"/>
        <gmr:ColInfo No="3" Unit="48" MarginA="2" MarginB="2"/>
      </gmr:Cols>
      <gmr:Rows DefaultSizePts="12.8">
        <gmr:RowInfo No="0" Unit="12.8" MarginA="0" MarginB="0" Count="14"/>
      </gmr:Rows>
      <gmr:Selections CursorCol="0" CursorRow="0">
        <gmr:Selection startCol="0" startRow="0" endCol="0" endRow="0"/>
      </gmr:Selections>
      <gmr:Cells>'''
      
    data = data + sheetHeader.split("\n")

    spaces = " " * 8
    
    for row in range(len(results)):
        for col in range(len(results[row])):
            if results[row][col] not in [[], ""]:
                cellData = spaces + \
                           '<gmr:Cell Col="%d" Row="%d" ' % (col, row)
                if type(results[row][col]) == type(1):
                    cellData += 'ValueType="30"'
                elif type(results[row][col]) == type(""):
                    if results[row][col][0] != '=':
                        cellData += 'ValueType="60"'
                elif type(results[row][col]) == type(0.0):
                    cellData += 'ValueType="40"'
                    
                cellData += ">" + str(results[row][col]) + "</gmr:Cell>\n"
                    
                data.append(cellData)
                    
    data += """      </gmr:Cells>
      <gmr:SheetLayout TopLeft="A1"/>
      <gmr:Solver TargetCol="-1" TargetRow="-1" ProblemType="1" Inputs=""/>
    </gmr:Sheet>""".split("\n")

    for ind in range(len(data)):
        if data[ind] and data[ind][-1] != "\n":
            data[ind] += "\n"
            
    return data

def extractCellInfo(cellLine):
    """Returns col, row, valuetype, value"""
    pattern = re.compile('.*Col="(\d*)".*Row="(\d*)".*ValueType="(\d*)".*\>(.*)\</gmr:Cell')
    m = pattern.match(cellLine)
    if m:
        return (int(m.group(1)), int(m.group(2)), int(m.group(3)), m.group(4))
    else:
        return (-1, -1, -1, None)

def cellName(row, col):
    """Return the Gnumeric name of this cell, given its 0-based row/col"""
    # TODO: This doesn't work beyond column YZ
    row += 1
    mod = col % 26
    div = col / 26
    ret = chr(ord("A") + mod)
    while div:
        col /= 26
        mod = col % 26
        div = col / 26
        ret = chr(ord("A") + mod) + ret
        tweak = 0
    return ret + str(row)
    
def composeRefCell(row, col, destcol, name):
    name = "'" + name + "'"        
    return '<gmr:Cell Col="%d" Row="%d">=%s!%s</gmr:Cell>\n' % \
           (col, row, name, cellName(row, destcol))


def addSummaryResultColumn(filename, name, coloff, warn=None, ignoring=True):
    data = gzip.open(filename).readlines()

    tofind = "<gmr:Name>" + name + "</gmr:Name>"
    ind = findSubStrIndex(data, tofind, 0, len(data))

    tofind = "<gmr:Cells>"
    resultsStart = findSubStrIndex(data, tofind, ind, len(data))
    resultsEnd = findSubStrIndex(data, "</gmr:Cells>", resultsStart, len(data))

    maxResultRow = 0
    for ind in range(resultsStart, resultsEnd):
        (col, row, valtype, value) = extractCellInfo(data[ind])
        if row > maxResultRow:
            maxResultRow = row
            
        
    tofind = "<gmr:Name>Summary</gmr:Name>"
    ind = findSubStrIndex(data, tofind, 0, len(data))

    tofind = "<gmr:Cells>"
    start = findSubStrIndex(data, tofind, ind, len(data))
    end = findSubStrIndex(data, "</gmr:Cells>", start, len(data))

    maxcol = 0
    maxrow = 0
    found = None
    for ind in range(start, end):
        (col, row, valtype, value) = extractCellInfo(data[ind])
        if value == name:
            found = col
            
        if col > maxcol:
            maxcol = col

        if row > maxrow:
            maxrow = row

    ignore = []
    if found != None:
        for ind in range(start, end):
            (col, row, valtype, value) = extractCellInfo(data[ind])
            if col != -1:
                #print "Good regexp"
                pass

            if col == found:
                if row == 0 or not ignoring:
                    del data[row]
                    end -= 1
                else:
                    ignore.append(row)
    else:
        #print "Didn't find", name, "maxcol =", maxcol
        found = maxcol + 1

    data[end:end] = [" " * 8 + '<gmr:Cell Col="%d" Row="0" ' % (found,) + \
                     'ValueType="60">%s</gmr:Cell>\n' % (name,)]

    if (maxResultRow > maxrow):
        namecol = getColumn(filename, name, 0)
        for i in range(maxrow + 1, maxResultRow + 1):
            data[end:end] = [" " * 8 + '<gmr:Cell Col="2" Row="%d" ' % (i,) +\
                             'ValueType="60">%s</gmr:Cell>\n' % (namecol[i],)]
            end += 1

    for i in range(1, maxResultRow + 1):
        if i not in ignore:
            data[end:end] = [composeRefCell(i, found, coloff, name)]
            end += 1

    outfile = gzip.open(filename, "w")
    for i in data:
        print >>outfile, i,
    outfile.close()

    if warn and maxResultRow > maxrow:
        return "Some usernames do not appear on summary sheet, check results."

        
def fixResults(filename, results):
    data = gzip.open(filename).readlines()
    rowLookup = {}
    rowLookup[0] = results[0]

    # Find the beginning of the summary sheet
    tofind = "<gmr:Name>Summary</gmr:Name>"
    ind = findSubStrIndex(data, tofind, 0, len(data))

    # Find the cell data for that sheet
    tofind = "<gmr:Cells>"
    start = findSubStrIndex(data, tofind, ind, len(data))
    end = findSubStrIndex(data, "</gmr:Cells>", start, len(data))

    max = 0
    for ind in range(start, end):
        (col, row, valtype, val) = extractCellInfo(data[ind])
        if col == 2:
            rowLookup[val] = row
            if row > max:
                max = row

    newResults = []
    for i in range(max + 1):
        newResults.append([])
    newResults[0] = results[0]
    
    for i in results[1:]:
        name = i[0]

        # Find which row name corresponds to within our range
        try:
            newResults[rowLookup[name]] = i
        except IndexError:
            print len(newResults), rowLookup[name]
            raise
        except KeyError:
            newResults += [i]

    return newResults

if __name__ == "__main__":
    results = extractResultsSheet("prac.gnumeric", "Sheet1")
    results = fixResults("014.gnumeric", results)

    mergeResultsSheet("014.gnumeric", results, "Lab Prac 1")
